<?php

namespace Drupal\moderation_scheduler;

use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Routing\RedirectDestinationInterface;
use Drupal\node\NodeListBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\content_moderation\ModeratedNodeListBuilder;
use Drupal\content_moderation\Entity\ContentModerationState as ContentModerationStateEntity;
use Drupal\content_moderation\Entity\ContentModerationStateInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityPublishedInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\content_moderation\Form\EntityModerationForm;
use Drupal\Core\Routing\RouteBuilderInterface;
use Drupal\workflows\Entity\Workflow;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Url;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Link;
use Drupal\Core\Render\Markup;
use Drupal\Component\Utility\Xss;
use Drupal\Core\Entity\Display\EntityViewDisplayInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Form\FormStateInterface;
use Drupal\node\Entity\NodeType;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Datetime\Entity\DateFormat;
use Drupal\Core\Extension\ModuleHandler;

/**
 * Defines a moderation scheduler interface.
 */
class ModerationSchedulerInterface {
  /**
   * Module handler service object.
   *
   * @var \Drupal\Core\Extension\ModuleHandler
   */
  protected $moduleHandler;

  /**
   * Constructs a SchedulerManager object.
   */
  public function __construct(ModuleHandler $moduleHandler) {
    $this->moduleHandler = $moduleHandler;
  }
    /**
    * Main function to publish scheduled node.
    *
    * @return array
    *   Array of published nodes.
    *
    * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
    */
    public function publishScheduled(){
        /** @var \Symfony\Component\EventDispatcher\EventDispatcherInterface $dispatcher */
        $dispatcher = \Drupal::service('event_dispatcher');
        //$this = ModerationSchedulerInterface::class;
        $node_storage = \Drupal::entityManager()->getStorage('node');
        $state = 'published';
        $default_language = \Drupal::languageManager()->getDefaultLanguage()->getId();

        // Get a date string suitable for use with entity query.
        $date = new DrupalDateTime();
        $date->setTimezone(new \DateTimezone(DateTimeItemInterface::STORAGE_TIMEZONE));
        $date = $date->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT);

        $languages = array_keys(\Drupal::languageManager()->getLanguages());

        $query = \Drupal::entityQuery('node')
            ->exists('field_scheduled_time')
            ->condition('field_scheduled_time.value', $date, '<=')
            ->latestRevision()
            ->sort('field_scheduled_time')
            ->sort('nid');
          // Disable access checks for this query.
          // @see https://www.drupal.org/node/2700209
          $query->accessCheck(FALSE);
          $nids = $query->execute();

          $nodes = $this->loadNodes($nids);

          if($nodes) {


            foreach ($nodes as $node) {

            $languages = $node->getTranslationLanguages();

            $translations = [];
            //If node has translation do logic to get latest updated translation revision
            if($languages) {

                $d_node_vid = $node_storage->getLatestTranslationAffectedRevisionId($node->id(), $default_language);

                $d_node = $node_storage->loadRevision($d_node_vid)->getTranslation($default_language);

                if($d_node->hasField('moderation_state')){
                    $old_state = $d_node->get('moderation_state')->value; 
                    $d_node->set('moderation_state', $state);
                } else {
                    $old_state = $d_node->status = 1 ? "published" : "unpublished";
                }

                if($d_node->hasField('langcode')){
                    $d_node->set('langcode', $default_language);
                }
                if($d_node->hasField('field_scheduled_time')){
                    $d_node->set('field_scheduled_time', NULL);
                }

                $d_node->setPublished(TRUE);

                $d_node->status = "1";

                    foreach ($languages as $language) {
                        $rev_lang = $language->getId();
                        $lang_name = $language->getName();
                        if($rev_lang != $default_language) {
                            //Remove the translation attached to latest revsision
                            $old_translation = $node->getTranslation($rev_lang);
                            if($d_node->hasTranslation($rev_lang)) {
                                try {
                                    $d_node->removeTranslation($rev_lang);                                
                                }
                                catch (\InvalidArgumentException $exception) {
                                    $old_translation = $node->getTranslation($rev_lang);
                                }
                            }
                            //Get latest translation revision by scheduled field
                            //Ensure old translation will not be lost if there is no new one
                            if($node->id()) {
                                $t_node = $this->get_revision_from_field($node->id(), $rev_lang) ? $this->get_revision_from_field($node->id(), $rev_lang) : $old_translation;                     
                            }
                            if($t_node) {
                                if($t_node->hasField('langcode')){
                                    $t_node->set('langcode', $rev_lang);
                                }
                                if($t_node->hasField('field_scheduled_time')){
                                    $t_node->set('field_scheduled_time', NULL);
                                }
                                if($t_node->hasField('moderation_state')){
                                    $t_node->set('moderation_state', $state);
                                }
                                $t_node->setPublished(TRUE);
                                $t_node->status = "1";

                                try {
                                    //If translation is not valid, throw exception.
                                    $translation = $d_node->addTranslation($rev_lang, $t_node->toArray())->save();
                                    $t_node->save();
                                    $t_result = [
                                       "lang" =>  $rev_lang. " - ".$lang_name." Translation (".$t_node->getTitle().")",
                                    ];
                                    $tranlsated = TRUE;
                                    array_push($translations, $t_result);
                                }
                                catch (\InvalidArgumentException $exception) {
                                    try {
                                        //Remove default language translation attached to translated node
                                        //If translation is not valid, throw exception.
                                        try {
                                            $t_node->removeTranslation($default_language);
                                            $translation = $t_node->addTranslation($default_language, $d_node->toArray())->save();
                                            $t_result = [
                                                "lang" =>  $rev_lang. " - ".$lang_name." Translation (".$d_node->getTitle().")",
                                             ];
                                            $tranlsated = TRUE;
                                            array_push($translations, $t_result);
                                        }
                                        catch (\InvalidArgumentException $exception) {
                                            continue;
                                        }
                                    }
                                    catch (\InvalidArgumentException $exception) {
                                        $t_title = $t_node->getTitle();
                                        $t_node->save();
                                    }
                                }
                            }
                        }
                    }
                } else {
                    $d_node_vid = $node_storage->getLatestRevisionId($node->id());

                    $d_node = $node_storage->loadRevision($d_node_vid);

                    if($d_node->hasField('moderation_state')){
                        $old_state = $d_node->get('moderation_state')->value; 
                        $d_node->set('moderation_state', $state);
                    } else {
                        $old_state = $d_node->status = 1 ? "published" : "unpublished";
                    }

                    if($d_node->hasField('langcode')){
                        $d_node->set('langcode', $default_language);
                    }
                    if($d_node->hasField('field_scheduled_time')){
                        $d_node->set('field_scheduled_time', NULL);
                    }

                    $d_node->setPublished(TRUE);

                    $d_node->status = "1";                
                }
                $d_title = $d_node->getTitle();
                //Save changes to default node revision
                $d_node->save();
                // Trigger the PUBLISH event so that modules can react after the node is
                // published.
                $event = new ModerationSchedulerEvent($d_node);
                $dispatcher->dispatch(ModerationSchedulerEvents::PUBLISH, $event);
                //Result log message
                $result[] = [
                                "node id" =>    $node->id(),
                                "old state" => $old_state,
                                "new state" => $state,
                                "lang" => $default_language. " - Original Translation (".$d_title.")",
                                "translations" => count($translations) >= 1 ? $translations : "no translation"
                            ];
            }
          } else {
            //If there isn't node with field_scheduled_time in default language
            //get iterating on languages enabled and field_scheduled_time revision table
            $languages = \Drupal::languageManager()->getLanguages();
            if(isset($languages)) {
                foreach ($languages as $language) {
                    $rev_lang = $language->getId();
                    $lang_name = $language->getName();
                    $translations = [];
                    $t_nodes = $this->get_revision_from_field(NULL, $rev_lang) ? $this->get_revision_from_field(NULL, $rev_lang) : [];

                    if(count($t_nodes) >= 1 && $rev_lang != $default_language) {
                        foreach ($t_nodes as $t_node) {
                            if($t_node) {
                                if($t_node->hasField('moderation_state')){
                                    $old_state = $t_node->get('moderation_state')->value; 
                                    $t_node->set('moderation_state', $state);
                                } else {
                                    $old_state = $t_node->status = 1 ? "published" : "unpublished";
                                }
                                $d_node_vid = $node_storage->getLatestTranslationAffectedRevisionId($t_node->id(), $default_language);

                                $d_node = $node_storage->loadRevision($d_node_vid)->getTranslation($default_language);  

                                if($t_node->hasField('langcode')){
                                    $t_node->set('langcode', $rev_lang);
                                }
                                if($t_node->hasField('field_scheduled_time')){
                                    $t_node->set('field_scheduled_time', NULL);
                                }
                                if($t_node->hasField('moderation_state')){
                                    $t_node->set('moderation_state', $state);
                                }
                                $t_node->setPublished(TRUE);
                                $t_node->status = "1";

                                try {
                                    //If translation is not valid, throw exception.
                                    $translation = $d_node->addTranslation($rev_lang, $t_node->toArray())->save();
                                    $t_node->save();
                                    $t_result = [
                                       "lang" =>  $rev_lang. " - ".$lang_name." Translation (".$t_node->getTitle().")",
                                    ];
                                    $tranlsated = TRUE;
                                    array_push($translations, $t_result);
                                }
                                catch (\InvalidArgumentException $exception) {
                                    try {
                                        //Remove default language translation attached to translated node
                                        //If translation is not valid, throw exception.
                                        try {
                                            $t_node->removeTranslation($default_language);
                                            $translation = $t_node->addTranslation($default_language, $d_node->toArray())->save();
                                            $t_result = [
                                                "lang" =>  $rev_lang. " - ".$lang_name." Translation (".$d_node->getTitle().")",
                                             ];
                                            $tranlsated = TRUE;
                                            array_push($translations, $t_result);
                                        }
                                        catch (\InvalidArgumentException $exception) {
                                            continue;
                                        }
                                    }
                                    catch (\InvalidArgumentException $exception) {
                                        $t_title = $t_node->getTitle();
                                        $t_node->save();
                                    }
                                }
                            // Trigger the PUBLISH event so that modules can react after the node is
                            // published.
                            $event = new ModerationSchedulerEvent($t_node);
                            $dispatcher->dispatch(ModerationSchedulerEvents::PUBLISH, $event);
                            $t_node->save();
                                //Result log message
                                $result[] = [
                                    "node id" =>    $t_node->id(),
                                    "old state" => $old_state,
                                    "new state" => $state,
                                    "lang" => $default_language. " - Original Translation (".$d_title.")",
                                    "translations" => count($translations) >= 1 ? $translations : "no translation"
                                ];
                            }
                        }
                    }
                }            
            }

          }
        //return result of scheduling
        return isset($result) ? $result : ["no scheduled content to publish"];
    }
    /**
     * Helper method to load latest revision of each node.
     *
     * @param array $nids
     *   Array of node ids.
     *
     * @return array
     *   Array of loaded nodes.
     *
     * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
     */
    public function loadNodes(array $nids) {
        $node_storage = \Drupal::entityManager()->getStorage('node');
        $nodes = [];
        // Load the latest revision for each node.
        foreach ($nids as $nid) {
          $vid = $node_storage->getLatestRevisionId($nid);
          $nodes[] = $node_storage->loadRevision($vid);
        }
        return $nodes;
    }
    /**
     * Helper method to load latest revision of each translation from scheduled field revisions.
     *
     * @param int $nid
     *   Integer node id.
     * 
     * @param string $langcode
     *   String langcode of node.
     * 
     * @return object or array of objects if no nid passed
     *   Object of loaded node revision or [Object node].
     *
     * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
     */
    public function get_revision_from_field($nid, $langcode) {
        $node_storage = \Drupal::entityManager()->getStorage('node');
        // Get a date string suitable for use with entity query.
        $date = new DrupalDateTime();
        $date->setTimezone(new \DateTimezone(DateTimeItemInterface::STORAGE_TIMEZONE));
        $date = $date->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT);
        $connection = \Drupal::database();
        if($nid) {
            //Query database to get revision id of latest node tranlsation scheduled
            $query = $connection->query("SELECT revision_id, langcode, field_scheduled_time_value, entity_id FROM {node_revision__field_scheduled_time} WHERE langcode = :langcode AND field_scheduled_time_value <= :date AND entity_id =:nid", [
                        ':langcode' => $langcode,
                        ':date' => $date,
                        ':nid' => $nid
                    ]);

            $revisions = $query->fetchAll();

                if($revisions) {
                    $revision = end($revisions);
                        //get revision id from the field
                        $revision_id = $revision->revision_id;
                        $vid = $revision->entity_id;
                        $schedule = $revision->field_scheduled_time_value;
                        $lang = $revision->langcode;
                        if($lang == $langcode && $schedule <= $date) {
                            //Load node revision with proper langcode and translation
                            $rev = $node_storage->loadRevision($revision_id)->getTranslation($langcode);
                            if($rev) {
                                $rev->set('langcode', $langcode);
                                return $rev;
                            }
                        }
                }
        } else {
            //If no nid Query database to get revision id of latest node tranlsation scheduled
            $query = $connection->query("SELECT revision_id, langcode, field_scheduled_time_value, entity_id FROM {node_revision__field_scheduled_time} WHERE langcode = :langcode AND field_scheduled_time_value <= :date", [
                        ':langcode' => $langcode,
                        ':date' => $date
                    ]);    
            $revisions = $query->fetchAll();
                if($revisions) {
                    $nodes = [];
                    foreach ($this->unique_multidim_array($revisions, 'entity_id') as $revision) {
                        //get revision id from the field
                        $revision_id = $revision->revision_id;
                        $vid = $revision->entity_id;
                        $schedule = $revision->field_scheduled_time_value;
                        $lang = $revision->langcode;      
                        if($lang == $langcode && $schedule <= $date) {
                            //Load node revision with proper langcode and translation
                            $rev = $node_storage->loadRevision($revision_id)->getTranslation($langcode);
                            if($rev) {
                                $rev->set('langcode', $langcode);
                                array_push($nodes, $rev);
                            }
                        }
                    }
                return count($nodes) >= 1 ? $nodes : NULL;
                }
        }
    }
    /**
     * Helper method to load unique_multidim_array by key.
     *
     * @param array [object node]
     *   array of nodes.
     *
     * @param string key
     *   string of key index.
     * 
     * @return array [object node]
     *   array of [object node].
     *
     */
    public function unique_multidim_array($array, $key) { 
        $temp_array = array(); 
        $i = 0; 
        $key_array = array(); 

        foreach($array as $val) { 
            if (!in_array($val->$key, $key_array)) { 
                $key_array[$i] = $val->$key; 
                $temp_array[$i] = $val; 
            } 
            $i++; 
        } 
        return $temp_array; 
    } 
    /**
     * Helper method to load latest revision of each translation from scheduled field revisions.
     *
     * @param string $field_name
     *   string of field_name.
     * 
     * 
     * @return bolean
     *   Bolean of result.
     *
     */
    public function clean_scheduled_time_field_revision($field) {

        $db = \Drupal::database();

        $query1 = $db->delete('node_revision__field_scheduled_time');
        $query2 = $db->delete('node__field_scheduled_time');
        $delete = $query1->execute();    
        $delete1 = $query2->execute();    

        if($delete >= 1) {
            return TRUE;   
        } else {
            return FALSE;
        }
    }
}